#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
from argparse import ArgumentParser, Namespace
from pathlib import Path
from typing import Any, Optional

from rich.table import Column, Table

from ...agents.coordinator import Coordinator
from ...testplan.schema import TestPlan
from ...utils.yaml import Dumper
from .. import YChaosCLIHook, YChaosTestplanInputSubCommand

__all__ = ["Attack"]


class Attack(YChaosTestplanInputSubCommand):
    """
    Agent Attack subcommand is used to perform the attack on the target. This
    subcommand requires a valid testplan to perforrm the required attack on
    the target system.
    """

    name = "attack"
    help = "YChaos Agent Attack Subcommand"

    @classmethod
    def build_parser(cls, parser: ArgumentParser) -> ArgumentParser:
        super(Attack, cls).build_parser(parser)
        parser.add_argument(
            "--attack-report-yaml",
            type=Path,
            help="File Path to store attack report in YAML format",
            default=None,
            metavar="path",
        )
        return parser

    def __init__(self, **kwargs):
        super(Attack, self).__init__(**kwargs)

        self.test_plan_path: Path = kwargs.pop("testplan")
        self.attack_report_yaml_path: Optional[Path] = kwargs.pop("attack_report_yaml")
        if self.attack_report_yaml_path and self.attack_report_yaml_path.is_dir():
            self.console.log(f"{self.attack_report_yaml_path} is not a valid file path")
            self.attack_report_yaml_path = None
        self.test_plan: Optional[TestPlan] = None
        self.coordinator: Optional[Coordinator] = None

    def validate_and_load_test_plan(self) -> int:
        self.test_plan = super(Attack, self).get_validated_test_plan(
            self.test_plan_path
        )
        return self._exitcode

    def configure_attack(self):
        self.coordinator = Coordinator(self.test_plan)

        class YChaosAgentAttackCLIHook(YChaosCLIHook):
            def __init__(self, app):
                super(YChaosAgentAttackCLIHook, self).__init__(app)

        class OnAttackStart(YChaosAgentAttackCLIHook):
            def __call__(self):
                self.console.log("Attack Started")

        class OnAttackCompleted(YChaosCLIHook):
            def __call__(self):
                self.console.log("Attack Ended")

        class OnAgentStart(YChaosAgentAttackCLIHook):
            def __call__(self, agent_name: str):
                self.console.log(f"Agent: {agent_name} - Started")

        class OnAgentTeardown(YChaosAgentAttackCLIHook):
            def __call__(self, agent_name: str):
                self.console.log(f"Agent: {agent_name} - Teardown started")

        class OnAgentStop(YChaosAgentAttackCLIHook):
            def __call__(self, agent_name: str):
                self.console.log(f"Agent: {agent_name} - Stopped")

        self.coordinator.register_hook("on_attack_start", OnAttackStart(self.app))
        self.coordinator.register_hook(
            "on_attack_completed",
            OnAttackCompleted(self.app),
        )
        self.coordinator.register_hook("on_each_agent_start", OnAgentStart(self.app))
        self.coordinator.register_hook(
            "on_each_agent_teardown",
            OnAgentTeardown(self.app),
        )
        self.coordinator.register_hook("on_each_agent_stop", OnAgentStop(self.app))

        self.coordinator.configure_agent_in_test_plan()
        table = Table(
            Column("Agent", style="green"),
            Column("Start Delay", justify="center", style="green"),
            Column("Start Time", style="green"),
            Column("End Time", style="green"),
            title="Configured Agents",
            show_header=True,
            header_style="bold magenta",
        )
        for configured_agent in self.coordinator.configured_agents:
            table.add_row(
                configured_agent.agent.config.name,
                str(configured_agent.agent.config.start_delay),
                str(configured_agent.start_time),
                str(configured_agent.end_time)
                if hasattr(configured_agent.agent.config, "duration")
                else "Unknown",
            )
        self.console.print(table)

    def print_all_errors(self):
        all_exceptions = self.coordinator.get_all_exceptions()
        for e in all_exceptions:
            try:
                raise e
            except Exception:
                self.console.print_exception()

    def dump_attack_report(self):
        if self.attack_report_yaml_path:
            import yaml

            with open(
                self.attack_report_yaml_path,
                "w",
            ) as fp:
                yaml.dump(
                    self.coordinator.generate_attack_report(),
                    fp,
                    default_flow_style=False,
                    sort_keys=False,
                    Dumper=Dumper,
                    indent=4,
                )
            self.console.log(f"Attack report stored at {self.attack_report_yaml_path}")

    @classmethod
    def main(cls, args: Namespace) -> Any:
        agent = cls(**vars(args))
        if agent._exitcode:
            return agent._exitcode
        if not agent.validate_and_load_test_plan():
            agent.configure_attack()
            assert agent.coordinator is not None

            agent.coordinator.start_attack()

            agent.print_all_errors()

            agent.dump_attack_report()

            if agent._exitcode or agent.coordinator.get_exit_status():
                agent.console.log("Attack failed")
                agent.set_exitcode(1)
        return agent._exitcode



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import collections
from typing import Callable, Dict, List


class InvalidEventHookError(KeyError):
    pass


class EventHook(object):
    @classmethod
    def CallableType(cls, *arg_types):
        """
        Returns an Alias of Callable that takes in `arg_types`
        """
        # IGNORE: `Callable' must be used as 'Callable[[arg, ...], result]`
        # The Argument types are casted to List before passing.
        return Callable[list(arg_types), None]

    __hook_events__: Dict[str, Callable[..., None]] = dict()
    """
    Lists the valid hooks (and corresponding valid Callable mapping) that can be registered for this
    particular object. The `register_hook()` method checks this list
    for the hooks that are being registered.
    """

    def __init__(self):
        """
        Initializes an event hook object
        """
        self.hooks: Dict[str, List[Callable]] = collections.defaultdict(list)

    def register_hook(self, event_name: str, hook: Callable) -> None:
        """
        Register a hook to be executed at a certain event `event_name`
        Args:
            event_name: Event name
            hook: A callable

        Returns:
            None
        """
        if event_name not in self.__hook_events__:
            raise InvalidEventHookError(event_name)

        self.hooks[event_name].append(hook)

    def execute_hooks(self, event_name: str, *args) -> None:
        """
        Execute all the hooks registered for a particular event
        Args:
            event_name: Event name
            *args: The arguments that should be passed to the hook

        Returns:
            None
        """
        if event_name not in self.__hook_events__:
            raise InvalidEventHookError(event_name)

        for hook in self.hooks[event_name]:
            try:
                if getattr(hook, "active", True):
                    hook(*args)
            except Exception as hook_error:  # nosec
                if getattr(hook, "raise_error", False):
                    raise hook_error



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms


class AgentError(Exception):
    def __init__(self, msg: str, **kwargs):
        self.msg = msg


class InsufficientPermissionError(AgentError):
    pass



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
from abc import ABC, abstractmethod

from ...testplan.schema import TestPlan
from ...utils.hooks import EventHook
from ..exceptions.executor_errors import YChaosTargetConfigConditionFailedError


class BaseExecutor(EventHook, ABC):
    """
    TargetExecutor defines the target where the agents are executed to
    test the resiliency of the system. A simple example of Target Executor is the
    [MachineTargetExecutor][ychaos.core.executor.MachineTargetExecutor.MachineTargetExecutor]
    which holds the program logic to execute the agents (with Coordinator) in Virtual Machines/BareMetals

    Each new Target Executor overrides the `execute()` method which defines "what" is to be done
    to execute the agents in a particular target environment.

    This class extends the [EventHook][ychaos.utils.hooks.EventHook] class, which
    implies each of the target executor can define its own events and the hooks that will
    be called during the trigger of an event.
    """

    __target_type__: str

    def __init__(self, testplan: TestPlan, *args, **kwargs):
        super(BaseExecutor, self).__init__()
        self.testplan = testplan
        self._validate_target_config()

    def _get_target_type(self):
        # To avoid circular import
        from ...testplan.attack import TargetType

        return TargetType(self.__target_type__)

    def _validate_target_config(self):
        target_config = self.testplan.attack.get_target_config()
        if self.testplan.attack.target_type != self._get_target_type():
            raise YChaosTargetConfigConditionFailedError("Target type mismatch")

        # Even though ideally this branch is never entered by the code (unless there is some issue with pydantic)
        if not isinstance(
            target_config, self._get_target_type().metadata.schema
        ):  # pragma: no cover
            raise YChaosTargetConfigConditionFailedError(
                "Target configuration is not processable for this executor"
            )

    @abstractmethod
    def execute(self) -> None:
        """
        Define "what" is to be done when the testplan consists
        the instruction to execute the agents in a particular
        target environment.

        Returns:
            None
        """
        pass  # Implement in Executors



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from pathlib import Path
from typing import Any, Union

import yaml

from .schema import TestPlan


class TestPlanValidator:
    """
    The Test plan validator class. Provides utility methods
    to validate the test plan files or raw data.
    """

    @classmethod
    def validate_file(cls, path: Union[str, Path]) -> None:
        """
        Validate a file (JSON/YAML) if the file content adheres to
        the testplan schema or not. This method on successful validation
        should not throw any errors.

        Raises:
            pydantic.ValidationError: If the content of the file does not follow the test plan schema
            FileNotFoundError: If the path passed to the method does not exist

        Args:
            path: A pathlike object

        Returns:
            None.
        """
        path = Path(path)
        with open(path, "r") as file:
            data = yaml.safe_load(file)

        cls.validate_data(data)

    @classmethod
    def validate_data(cls, data: Any) -> None:
        """
        Validate raw data (e.g. dictionary) whether it follows the Test plan schema.
        This method on successful validation should not throw any errors.

        Raises:
            pydantic.ValidationError: If the data does not follow the test plan schema

        Args:
            data: A data object (dictionary)

        Returns:
            None
        """
        TestPlan.validate(data)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

import logging
import os
import time
from logging.handlers import QueueHandler, QueueListener
from queue import Queue
from typing import Optional, Union

from .settings import DevSettings, ProdSettings, Settings
from .utils.logging import StructLogger


class AppLogger:

    __instance: Optional[StructLogger] = None
    __log_queue: Queue = Queue(maxsize=1000)
    _listener: Optional[QueueListener] = None

    def __init__(self):
        """
        Configures the root logger to custom logging structure and adds log handlers based on the env
        """
        settings: Union[DevSettings, ProdSettings] = Settings.get_instance()

        logging.setLoggerClass(StructLogger)
        self.__class__.__instance = logging.getLogger(settings.PROG)
        self.__class__.__instance.setLevel(logging.DEBUG)

        formatter = logging.Formatter(
            (
                f"%(asctime)s [%(levelname)-s] host={os.uname()[1]} application=%(name)s module=%(module)s "
                f"method=%(funcName)s line=%(lineno)d %(message)s"
            )
        )
        formatter.converter = time.gmtime
        formatter.datefmt = "%Y-%m-%d %H:%M:%S"

        if settings.CONFIG == "prod" and settings.LOG_FILE_PATH is not None:
            queue_handler = QueueHandler(self.__class__.__log_queue)

            file_handler = logging.FileHandler(settings.LOG_FILE_PATH, "w")
            file_handler.setLevel(logging.DEBUG)
            file_handler.setFormatter(formatter)

            self.__class__.__instance.addHandler(queue_handler)
            self.__class__._listener = QueueListener(
                self.__class__.__log_queue, file_handler
            )

    @classmethod
    def get_logger(cls, name: str):
        if cls.__instance is None:
            cls()

        assert isinstance(cls.__instance, StructLogger)
        return cls.__instance.getChild(name)

    @classmethod
    def start(cls):
        if cls._listener:
            cls._listener.start()

    @classmethod
    def stop(cls):
        if cls._listener:
            cls._listener.stop()



#!/usr/bin/env python

#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

"""
Package setup file for python module ychaos
"""
# WARNING: Do not modify this file unless you know exactly what you are doing.
#
# Note: If you believe you know exactly what you are doing, you are most likely wrong
#
# This file is an executable python script that will be included in the sdist package, this file
# is used to obtain package metadata and install the package on other computer systems.   As a
# result, it must be able to execute without generating errors or execeptions in an environment with
# only the python interpreter and nothing else.
#
# This file should never do any of the following:
#  - Import Python modules that are not included with the Python standard library
#  - Import any code that is contained in the package
#  - Execute external commands (do not use the subprocess module, os.system, os.popen*)
#  - Access or use any file that is not included in this package
#  - Call the sys.exit() or exit() functions unless there is an error that will prevent the package from installing.
#  - Generate any output to stdout (do not use the print function or call or use any code that uses it)
import os
import setuptools


def scripts():
    """
    Get the scripts in the "scripts" directory

    Returns
    list
        List of filenames
    """
    script_list = []
    if os.path.isdir("scripts"):
        for item in os.listdir("scripts"):
            filename = os.path.join("scripts", item)
            if os.path.isfile(filename):
                script_list.append(filename)
    return script_list


if __name__ == "__main__":
    # We're being run from the command line so call setup with our arguments
    setuptools.setup(
        scripts=scripts(),
        packages=setuptools.find_packages(where="src"),
        package_dir={"": "src"},
        test_suite="tests",
        package_data={"": ["*.json"]},
    )



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import math
import shutil
from pathlib import Path
from queue import LifoQueue

from pydantic import Field

from ..agent import (
    Agent,
    AgentMonitoringDataPoint,
    AgentPriority,
    TimedAgentConfig,
)
from ..utils.annotations import log_agent_lifecycle


class DiskFillConfig(TimedAgentConfig):
    """
    Defines the Disk Fill configuration to consume the disk space. The framework will
    fill the disk space of a directory defined in `partition` and fill only a percentage of available
    space, as defined in the `partition_pct` attribute. By default,
    `partition_pct` is set to 100, implying the complete disk space of that partition will be filled.
    """

    name = "disk_fill"
    description = "This agent fills up disk space."

    priority = AgentPriority.MODERATE_PRIORITY

    partition: Path = Field(
        description="The Filepath of directory or partition to fill",
        default=Path("/etc/"),
        examples=["/etc/tmp", "/home/tmpuser"],
    )

    partition_pct: float = Field(
        default=80,
        description=(
            "Percentage of the disk partition to fill"
            "This will fill a percentage of the disk space on the partition"
        ),
        gt=0,
        le=100,
    )

    max_file_size: int = Field(
        default=(1024 * 1024 * 1024 * 20),  # Max file size is 20GB,
        description=(
            "Maximum size of each disk fill file. If the size to be filled exceeds this max file size, multiple"
            "disk fill files will be used"
        ),
        gt=1024,
    )

    disk_fill_dir: str = Field(
        default="ychaos_diskfill",
        description=(
            "Name of the temporary directory in which to store the disk fill files."
            "This will be the path relative to the partition to fill"
        ),
    )

    def effective_disk_to_fill(self) -> int:
        """
        Calculates the disk space that needs to be filled based on the available space in the partition and
        the percentage of partition to fill partition_pct.
        Returns:
            disk space to fill in given partition.
        """
        stat = shutil.disk_usage(self.partition)
        partition_size_available = stat.free
        return math.floor(self.partition_pct / 100 * partition_size_available)


class DiskFill(Agent):
    def monitor(self) -> LifoQueue:
        super(DiskFill, self).monitor()
        available_space = shutil.disk_usage(self.config.partition).free

        self._status.put(
            AgentMonitoringDataPoint(
                data=dict(
                    disk_space_to_fill=self.config.effective_disk_to_fill(),
                    disk_free_space=available_space,
                ),
                state=self.current_state,
            )
        )
        return self._status

    @log_agent_lifecycle
    def setup(self) -> None:
        super(DiskFill, self).setup()

    @log_agent_lifecycle
    def run(self) -> None:
        super(DiskFill, self).run()
        size = self.config.effective_disk_to_fill()
        if size <= 0:
            return
        disk_fill_dir = Path(self.config.partition / self.config.disk_fill_dir)
        disk_fill_dir.mkdir(parents=True, exist_ok=True)

        space_remaining = size
        index = 0
        while space_remaining > 0:
            if self.stop_async_run:
                break
            tmp = space_remaining
            cur_file_size = self.config.max_file_size
            space_remaining -= self.config.max_file_size
            if space_remaining <= 0:
                cur_file_size = tmp
            with open(disk_fill_dir / f"filler{index}.txt", "wb") as f:
                f.seek(cur_file_size - 1)
                f.write(b"\0")
            index += 1

    @log_agent_lifecycle
    def teardown(self) -> None:
        super(DiskFill, self).teardown()
        tmp_dir = self.config.partition / self.config.disk_fill_dir
        if tmp_dir.exists():
            shutil.rmtree(tmp_dir)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from abc import abstractmethod
from typing import Any, Dict


class YChaosCLIError(Exception):

    exitcode = 1

    def __init__(self, app, message, **kwargs):
        self.app = app
        self.message: str = message
        self.attrs: Dict[str, Any] = kwargs

    @abstractmethod
    def handle(self) -> None:
        """
        Handle is the method that is called during teardown to handle a particular type of error
        Any subcommand can raise a sub class of YChaosCLIError and forget about the exception.
        The main teardown module takes responsibility of calling the handle method

        This can be used to print message of exception or handle the panic in a suitable way
        """
        if self.app.is_debug_mode():
            self.app.console.print_exception(extra_lines=2)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import functools
import shlex
import sys
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Dict, List, Optional, Tuple, Union

from pydantic import AnyHttpUrl, Field, PositiveInt, SecretStr, validator

from ...utils.builtins import AEnum, BuiltinUtils
from .. import SchemaModel, SystemState
from ..common import Secret
from .plugins.metrics import (
    MultipleConditionalsMetricsVerificationCriteria,
    StateBoundMetricsVerificationCriteria,
)


class PythonModuleVerification(SchemaModel):
    """
    The Python Module verification. Runs an custom python script given
    the path of the script. The executable of the script is set to `sys.executable` by
    default which can be overridden by a custom executable. The script can also be passed
    a list of arguments that are provided in the arguments argument which are passed to
    the script by separating each element of the argument by space.

    The user of this plugin takes the entire responsibility of the script executed. YChaos
    is only responsible of running the script and collecting the exitcode of the script
    with which the state of the system is verified/monitored.
    """

    path: Path = Field(..., description="The absolute path of the python script")
    executable: str = Field(
        default=sys.executable,
        description="The python shell to be used for executing the script",
    )
    arguments: List[str] = Field(
        default=list(),
        description="List of arguments to be sent to the script. The arguments in the list will be sent to the script space separated",
    )

    def safe_arguments(self):
        return [shlex.quote(x) for x in self.arguments]


class HTTPRequestSchema(SchemaModel):
    """
    The Base class for Plugin that make a HTTP Network call.
    """

    method: str = Field(default="GET", description="HTTP method to be used")
    headers: Dict[str, Union[SecretStr, Secret]] = Field(
        default=dict(), description="Headers to be sent with the request"
    )
    params: Dict[str, str] = Field(
        default=dict(), description="Query params to be sent with the request"
    )
    verify: bool = Field(
        default=True, description="Verify the target URL SSL certificates"
    )

    # Authentication
    basic_auth: Optional[Tuple[str, Union[SecretStr, Secret]]] = Field(
        default=None, description="Basic Auth authentication for the HTTP call"
    )
    bearer_token: Optional[Union[SecretStr, Secret]] = Field(
        default=None, description="Bearer token authentication for the HTTP call"
    )

    # Certificate
    cert: Tuple[Path, Path] = Field(
        default=None,
        description="The certificate to be sent for HTTP call. The tuple should contain Certificate and Key File path",
    )

    timeout: int = Field(
        default=10000,
        description="Timeout in milliseconds at which the HTTP requests will timeout",
        gt=0,
    )

    _validate_method = validator("method", pre=True, allow_reuse=True)(
        BuiltinUtils.Request.validate_method
    )


class HTTPRequestVerification(HTTPRequestSchema):
    """
    Makes requests to the specified endpoints and verifies that
    the service is returning a valid response along with verifying the
    response time of the service being less than `latency` field.

    The user can also specify a `count` attribute requesting the tool to make
    `count` number of requests to the same endpoints.
    """

    count: int = Field(
        default=1,
        description="Number of HTTP calls to be sent to each of the URL",
        ge=1,
    )
    latency: int = Field(
        default=50,
        description="Expected Latency in ms. A latency below this value indicates successful verification",
        ge=1,
    )

    status_codes: List[int] = Field(
        default_factory=functools.partial(list, range(200, 300)),
        description="The list of status code which will be the comparison factor for the HTTP responses",
    )

    urls: List[AnyHttpUrl] = Field(
        default=list(), description="List of HTTP/s URLs to be requested"
    )


class SDv4Verification(SchemaModel):
    """
    The SDV4VerificationPlugin offers the user of YChaos to configure
    a 3rd party Screwdriver Job that is triggered remotely by YChaos and upon the
    successful completion of that SDv4 job, the verification is marked as successful.

    If the SDv4 job fails/is aborted, the verification is marked as failure.

    [Know more about Screwdriver CI/CD](https://screwdriver.cd/)
    """

    pipeline_id: int = Field(
        ..., description="SDv4 pipeline ID", examples=[123456, 1041241]
    )

    job_name: str = Field(
        ...,
        description="Job name in the pipeline",
        examples=["test_validation", "state_verification"],
    )

    sd_api_url: AnyHttpUrl = Field(..., description="SDv4 API URL")
    sd_api_token: Union[SecretStr, Secret] = Field(
        ...,
        description="The Screwdriver pipeline/user access token to be able to start the jon in the pipeline",
    )

    job_timeout: PositiveInt = Field(default=3600, description="Job Timeout in seconds")


class NoOpConfig(SchemaModel):
    pass


class OpenTSDBVerification(HTTPRequestSchema):
    """
    The OpenTSDB Verification Plugin gets the metrics from an OpenTSDB server and compares it with the
    provided comparison parameters in the testplan. If the condition passes
    """

    url: AnyHttpUrl = Field(
        ..., description="The OpenTSDB server URL to get the metrics from"
    )

    # Override method to change default value
    method: str = Field(default="POST", description="HTTP method to be used")

    query: Dict[str, Any] = Field(
        default=dict(), description="The OpenTSDB query sent to the server."
    )

    criteria: List[MultipleConditionalsMetricsVerificationCriteria] = Field(
        default=list(),
        description=(
            "Metrics verification criteria without state information."
            "All the criteria part of this list must pass for the verification to be successful"
        ),
    )

    state_bound_criteria: List[StateBoundMetricsVerificationCriteria] = Field(
        default=list(),
        description=(
            "Metrics verification criteria with state."
            "All the criteria part of this list must pass for the verification to be successful"
        ),
    )

    @validator("state_bound_criteria", pre=True, always=True)
    def _criteria_validation(cls, v, values):
        # The input must contain at least one of "criteria" or "state_bound_criteria"
        if not v and not values.get("criteria", list()):
            raise ValueError(
                "Either criteria or state_bound_criteria must be present for this configuration."
            )
        return v

    _validate_method = validator("method", pre=True, allow_reuse=True)(
        BuiltinUtils.Request.validate_method
    )


class VerificationType(AEnum):
    """
    Defines the Type of plugin to be used for verification.
    """

    # The metadata object will contain the following attributes
    # 1. schema : The Schema class of the VerificationType

    PYTHON_MODULE = "python_module", SimpleNamespace(schema=PythonModuleVerification)
    HTTP_REQUEST = "http_request", SimpleNamespace(schema=HTTPRequestVerification)
    SDV4_VERIFICATION = (
        "sdv4",
        SimpleNamespace(schema=SDv4Verification),
    )
    OPENTSDB_VERIFICATION = "tsdb", SimpleNamespace(schema=OpenTSDBVerification)

    # For Testing purpose, cannot be used by users.
    NOOP = "noop", SimpleNamespace(schema=NoOpConfig)


class VerificationConfig(SchemaModel):
    """
    The verification configuration that is executed during some state of the system
    to verify the system is in a favorable conditions or not.
    """

    delay_before: float = Field(
        default=0,
        description="delay (in ms) to be introduced before running this plugin",
    )

    delay_after: float = Field(
        default=0,
        description="delay (in ms) to be introduced after running this plugin",
    )

    states: Union[SystemState, List[SystemState]] = Field(
        ...,
        description="A system state or a list of system states in which this verification plugin should be executed.",
    )

    type: VerificationType = Field(..., description="The verification type to be used.")

    strict: bool = Field(
        default=True,
        description=(
            "Setting this value to false implies the overall verification"
            " does not fail because of the failure of this test."
        ),
    )

    config: Dict[str, Any] = Field(
        ..., description="The verification type configuration"
    )

    def get_verification_config(self):
        return self.type.metadata.schema(**self.config)

    @validator("states", pre=True)
    def _parse_states_to_list(cls, v):
        """
        Parses the state object to List of states to keep
        the access consistent
        Args:
            v: SystemState object/ List of SystemState

        Returns:
            List of SystemState
        """
        return BuiltinUtils.wrap_if_non_iterable(v)

    @validator("config", pre=True)
    def _parse_plugin_configuration(cls, v, values):
        """
        Validates the plugin configuration to match with the
        mapper class of the plugin.
        Args:
            v: Plugin configuration
            values: verification configuration

        Returns:
            Parsed mapper object
        """
        if "type" in values:
            return VerificationType(values["type"]).metadata.schema(**v)
        else:
            return v



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from argparse import ArgumentParser

from ...utils.argparse import SubCommandParsersAction
from .. import YChaosSubCommand
from .validate import TestPlanValidatorCommand

__all__ = ["TestPlan"]


class TestPlan(YChaosSubCommand):
    """
    The testplan sub-command. Provides commands to operate on testplan.
    See Testplan documentation for more details.
    """

    name = "testplan"
    help = "sub command for test plan operations"

    @classmethod
    def build_parser(cls, parser: ArgumentParser) -> ArgumentParser:
        test_plan_command_subparser = parser.add_subparsers(
            action=SubCommandParsersAction, dest=f"_cmd.{cls.name}"
        )

        test_plan_command_subparser.add_parser(
            name=TestPlanValidatorCommand.name, cls=TestPlanValidatorCommand
        )

        return parser



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

# TODO: PingDisable Agent
import warnings
from queue import LifoQueue

from ..agent import (
    Agent,
    AgentMonitoringDataPoint,
    AgentPriority,
    TimedAgentConfig,
)
from ..utils.annotations import log_agent_lifecycle
from ..utils.sysctl import SysCtl


class PingDisableConfig(TimedAgentConfig):
    name = "ping_disable"
    description = "This agent disables the system to respond to ping/icmp packets"

    priority = AgentPriority.MODERATE_PRIORITY
    is_sudo = True


class PingDisable(Agent):

    sysctl_var = "net.ipv4.icmp_echo_ignore_all"

    def monitor(self) -> LifoQueue:
        super(PingDisable, self).monitor()
        self._status.put(
            AgentMonitoringDataPoint(
                data=dict(),  # TODO: Ping localhost and add monitoring data
                state=self.current_state,
            )
        )
        return self._status

    @log_agent_lifecycle
    def setup(self) -> None:
        super(PingDisable, self).setup()
        self.preserved_state.is_ping_disabled = int(SysCtl.get(self.sysctl_var)) == 1

    @log_agent_lifecycle
    def run(self) -> None:
        super(PingDisable, self).run()
        if self.preserved_state.is_ping_disabled:
            warnings.warn(
                "ICMP ignore is already turned on. "
                "Running this agent will be a no-operation"
            )
        else:
            SysCtl.set(self.sysctl_var, b"1")

    @log_agent_lifecycle
    def teardown(self) -> None:
        super(PingDisable, self).teardown()
        if not self.preserved_state.is_ping_disabled:
            SysCtl.set(self.sysctl_var, b"0")



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

import importlib
import warnings
from typing import Any, Optional, Tuple

from pydantic import validate_arguments


class DependencyUtils:
    """
    DependencyUtils provides utility methods to handle and import optional dependencies
    in the YChaos package.
    """

    @classmethod
    def import_module(
        cls, name: str, message: str = None, raise_error: bool = True, warn: bool = True
    ) -> Optional[Any]:
        """
        Calling this method with a module name is similar to calling
        `import ...`. This can be used to import optional dependencies in the package.

        Args:
            name: Module name
            message: Error message to be printed on console when the import fails
            raise_error: Raise an error if the import fails
            warn: Raise warning if the import fails

        Raises:
            ImportError: when `raise_error` is True and the module is not present

        Returns:
            Optional Module
        """
        try:
            module = importlib.import_module(name)
        except ImportError as import_error:
            if not message:
                message = f"Dependency {name} is not installed."

            if warn:
                warnings.warn(message)

            if raise_error:
                raise ImportError(message) from None
            else:
                return None

        return module

    @classmethod
    @validate_arguments
    def import_from(
        cls,
        module_name: str,
        attrs: Tuple[str, ...],
        message: str = None,
        raise_error: bool = True,
        warn: bool = True,
    ) -> Tuple[Any, ...]:
        """
        Calling this method with a module and an attribute is similar to calling
        `from ... import ...`. This can be used to import optional dependency in the package.

        Examples:

            ```python
            from ychaos.utils.dependency import DependencyHandler
            BaseModel, Field = DependencyHandler.import_from("pydantic", ("BaseModel", "Field"))
            ```

            The above code snippet is same as
            ```python
            from pydantic import BaseModel, Field
            ```

        Args:
            module_name: Valid Python Module name
            attrs: Tuple of attribute names from the module
            message: message to be printed in case of an error
            raise_error: Raise an error if the import fails
            warn: Throw a warning if the import fails

        Raises:
            ImportError: when `raise_error` is true and `attr_name` cannot be imported from the `module_name`

        Returns:
            An attribute from module_name if exists, None otherwise
        """
        module = cls.import_module(
            name=module_name, message=message, raise_error=raise_error, warn=warn
        )

        if not module:
            return (None,) * len(attrs)
        else:
            _attr_list = list()
            for _attr_name in attrs:
                try:
                    attr = getattr(module, _attr_name)
                    _attr_list.append(attr)
                except AttributeError as attr_error:
                    if not message:
                        message = f"cannot import {_attr_name} from {module_name}"

                    if warn:
                        warnings.warn(message)
                    if raise_error:
                        raise ImportError(message) from None
                    else:
                        _attr_list.append(None)
            return tuple(_attr_list)



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms
import getpass
import re
from enum import Enum
from pathlib import Path
from types import SimpleNamespace
from typing import Any, Dict, List, Optional, Union

from pydantic import Field, FilePath, SecretStr, validator

from ..agents.index import AgentType
from ..utils.builtins import FQDN, AEnum
from . import SchemaModel
from .common import Secret


class TargetDefinition(SchemaModel):
    report_dir: Path = Field(
        default=Path("./"),
        description="The report directory to store execution files. Defaults to current directory",
    )


class SSHConfig(SchemaModel):
    user: str = Field(
        default=getpass.getuser(),
        description="The login user for SSH connection. Defaults to current user.",
    )
    private_key: Optional[Path] = Field(
        default=None,
        description="The private key file that will be used to login to the hosts.",
    )
    password: Union[SecretStr, Secret] = Field(
        default=None,
        description="The password that will be used to login to the hosts.",
    )


class MachineTargetDefinition(TargetDefinition):
    """
    Represents the configuration when the target is a Virtual machine
    or Baremetal. To use this as a target, `target_type` should be equal to
    `machine`

    Attributes:
        blast_radius: The percentage of targets to be attacked. **This is a required field**
        ssh_config: The SSH Configuration to be used while logging into the hosts. See [SSHConfig][ychaos.testplan.attack.SSHConfig]
        hostnames: List of hosts as targets to run the agents on. These should be valid FQDNs.
        hostpatterns: List of Host patterns with a single number range within the pattern
        hostfiles:
            List of files containing hostnames separated by a newline
            The path provided can be an absolute path or a relative path from which the tool is invoked.
        exclude:
            List of hosts to be always excluded out of the attack.
            The filtering criteria will always exclude the hosts in this list

    Warning:
        Testplan validator will not be able to validate each of the host entries if the targets are provided
        using `hostfiles`. The tool will dynamically validate each of the entry while reading the files.
    """

    blast_radius: int = Field(
        ..., description="The percentage of targets to be attacked", ge=0, le=100
    )

    ssh_config: SSHConfig = Field(
        ..., description="The configuration used to SSH to the target machines."
    )

    hostnames: List[FQDN] = Field(
        default=list(),
        description="List of hosts as targets to run the agents on. These should be valid FQDNs.",
        examples=[
            ["myhost01.yahoo.com", "myhost02.yahoo.com", "mockhost.web.fe.yahoo.com"],
        ],
    )

    hostpatterns: List[str] = Field(
        default=list(),
        description="List of Host patterns with a single number range within the pattern",
        examples=[
            ["myhost[12-34].yahoo.com", "hostpattern[00-10].mock.yahoo.com"],
        ],
    )

    hostfiles: List[FilePath] = Field(
        default=list(),
        description=(
            "List of files containing hostnames separated by a newline"
            "The path provided can be an absolute path or a relative path from which the tool is invoked."
            "Note that the testplan will not validate each file during static validation."
        ),
        examples=[
            ["/home/awesomeuser/hostlist.txt", "/home/awesomeuser/tmp/inventory.txt"]
        ],
    )

    exclude: List[FQDN] = Field(
        default=list(),
        description=(
            "List of hosts to be always excluded out of the attack."
            "The filtering criteria will always exclude the hosts in this list"
        ),
    )

    def iterate_hostfiles(self):
        for file in self.hostfiles:
            for host in file.read_text().strip().splitlines():
                yield FQDN(host)

    def iterate_hostpattern(self):
        for v in self.hostpatterns:
            match = re.search(r"\[((\d+)-(\d+))\]", v)
            if match is None:
                yield FQDN(v)

            else:
                range_start = match.group(2)
                range_end = match.group(3)

                for num in range(int(range_start), int(range_end) + 1):
                    yield FQDN(
                        v[: match.start()]
                        + str(num).zfill(len(range_start))
                        + v[match.end() :]
                    )

    def expand_hostpatterns(self) -> List[FQDN]:
        expanded_list = list()
        for hostname in self.iterate_hostpattern():
            expanded_list.append(hostname)

        return expanded_list

    def expand_hostfiles(self):
        expanded_list = list()
        for host in self.iterate_hostfiles():
            expanded_list.append(host)
        return expanded_list

    def get_effective_hosts(self):
        return list(
            set(
                self.expand_hostpatterns() + self.expand_hostfiles() + self.hostnames
            ).difference(set(self.exclude))
        )

    @validator("hostpatterns", pre=True, each_item=True)
    def validate_hostpatterns(cls, v):
        match = re.search(r"\[((\d+)-(\d+))\]", v)
        if match is None:
            FQDN(v)

        else:
            range_start = match.group(2)
            range_end = match.group(3)

            for num in range(int(range_start), int(range_end) + 1):
                FQDN(
                    v[: match.start()]
                    + str(num).zfill(len(range_start))
                    + v[match.end() :]
                )
        return v


class SelfTargetDefinition(TargetDefinition):
    """
    Represents the configuration when the target is your local machine
    . To use this as a target, `target_type` should be equal to
    `self`.
     You can avoid specifying `target_config` for this target type.
    """

    pass


class TargetType(AEnum):

    SELF = "self", SimpleNamespace(schema=SelfTargetDefinition)
    # The machine running the toolkit registers itself as the target to run all the agents

    MACHINE = "machine", SimpleNamespace(schema=MachineTargetDefinition)
    # Registers some virtual machines, Baremetals etc.


class AttackMode(Enum):
    """
    Defines the type of execution mode for executing the configured Agents
    """

    CONCURRENT = "concurrent"
    SEQUENTIAL = "sequential"


class AgentExecutionConfig(SchemaModel):

    type: AgentType = Field(
        ...,
        description=(
            "Defines the agent type to be executed on the target. "
            "The configuration of agent is determined by this attribute"
        ),
        examples=["cpu_burn", "no_op", "ping_disable"],
    )

    config: Dict[Any, Any] = Field(
        default=dict(),
        description="The Agent configuration for a particular agent type",
    )

    def get_agent_config(self):
        return self.type.metadata.schema(**self.config)

    @validator("config", pre=True)
    def _parse_agent_configuration(cls, v, values):
        if "type" in values:
            return AgentType(values["type"]).metadata.schema(**v)
        else:  # pragma: no cover
            return v


class AttackConfig(SchemaModel):

    target_type: TargetType = Field(
        ...,
        description=(
            "Defines the target type which will be used as a target for the execution of agents. "
            "The target_type configuration is determined the type mentioned in this attribute"
        ),
        examples=["self", "machine"],
    )

    target_config: Dict[Any, Any] = Field(
        default=dict(),
        description=(
            "Defines the targets for running the agents."
            "This can be set to null to imply that the framework should run all the agents within the same system where the executor has been invoked"
        ),
    )

    mode: AttackMode = Field(
        default=AttackMode.SEQUENTIAL,
        description="Define the execution mode for the attack",
    )

    agents: List[AgentExecutionConfig] = Field(
        default=list(),
        description=(
            "List of agents to be executed on the Target. "
            "Each of the item of execution configuration will infer a type of agent and a configuration of the agent"
        ),
        min_items=1,
    )

    def get_target_config(self):
        return self.target_type.metadata.schema(**self.target_config)

    @validator("target_config", pre=True)
    def _parse_target_configuration(cls, v, values):
        if "target_type" in values:
            return TargetType(values["target_type"]).metadata.schema(**v)
        else:  # pragma: no cover
            return v



#  Copyright 2021, Yahoo
#  Licensed under the terms of the Apache 2.0 license. See the LICENSE file in the project root for terms

from argparse import ArgumentParser, Namespace
from io import StringIO
from pathlib import Path
from typing import Any, Optional

from rich.markdown import Markdown

from . import YChaosSubCommand

__all__ = ["Manual"]


class Manual(YChaosSubCommand):
    """
    Used to print the manual for the entire CLI command. Can be used to
    auto-generate Markdown docs for the YChaos CLI by passing the `-f/--file`
    argument. In case, a file is not passed, the markdown is printed on
    the console.
    """

    name = "manual"
    help = "Print the manual for YChaos CLI"

    def __init__(self, **kwargs):
        super(Manual, self).__init__(**kwargs)

        self.file: Optional[Path] = kwargs.pop("file", None)

    @classmethod
    def build_parser(cls, parser: ArgumentParser) -> ArgumentParser:
        parser.add_argument(
            "-f",
            "--file",
            type=Path,
            help="Print YChaos CLI Manual to a file",
            required=False,
            metavar="path",
        )

        return parser

    def do_print_manual_entry(self):
        tempfile = StringIO()
        for cmd, msg in self.app.manual_entry().items():
            tempfile.write(f"{'#' * len(cmd.split())} {cmd}")
            tempfile.write("\n")
            tempfile.write("```\n")
            tempfile.write(msg)
            tempfile.write("\n```\n")

        if self.file is not None:
            try:
                self.file.write_text(tempfile.getvalue())
                self.console.log(
                    f"{self.app.settings.PROG} manual entry written to {self.file}"
                )
            except FileNotFoundError as file_not_found_error:
                self.set_exitcode(1)
                self.console.print(
                    f":mag: {self.file} [italic]not found[/italic]",
                    style="indian_red",
                )
            except IsADirectoryError as is_a_directory_error:
                self.set_exitcode(1)
                self.console.print(
                    f":file_folder: The input path ({self.file}) is a directory"
                )
        else:
            self.console.print(Markdown(tempfile.getvalue()))

    @classmethod
    def main(cls, args: Namespace) -> Any:  # pragma: no cover
        manual = cls(**vars(args))
        manual.do_print_manual_entry()

        return manual._exitcode
